@@ -4,8 +4,8 @@
 /* ************************************************************************** */
 /* *                                                                        * */
 /* * project   : libmng                                                     * */
-/* * file      : libmng_read.c             copyright (c) 2000-2003 G.Juyn   * */
-/* * version   : 1.0.6                                                      * */
+/* * file      : libmng_read.c             copyright (c) 2000-2004 G.Juyn   * */
+/* * version   : 1.0.8                                                      * */
 /* *                                                                        * */
 /* * purpose   : Read logic (implementation)                                * */
 /* *                                                                        * */
@@ -81,6 +81,18 @@
 /* *             1.0.6 - 08/17/2003 - G.R-P                                 * */
 /* *             - added conditionals around non-VLC chunk support          * */
 /* *                                                                        * */
+/* *             1.0.7 - 03/10/2004 - G.R-P                                 * */
+/* *             - added conditionals around openstream/closestream         * */
+/* *                                                                        * */
+/* *             1.0.8 - 04/08/2004 - G.Juyn                                * */
+/* *             - added CRC existence & checking flags                     * */
+/* *             1.0.8 - 04/11/2004 - G.Juyn                                * */
+/* *             - added data-push mechanisms for specialized decoders      * */
+/* *             1.0.8 - 07/06/2004 - G.R-P                                 * */
+/* *             - defend against using undefined closestream function      * */
+/* *             1.0.8 - 07/28/2004 - G.R-P                                 * */
+/* *             - added check for extreme chunk-lengths                    * */
+/* *                                                                        * */
 /* ************************************************************************** */
 
 #include "libmng.h"
@@ -115,8 +127,12 @@ mng_retcode mng_process_eof (mng_datap pData)
   {
     pData->bEOF = MNG_TRUE;            /* now we do! */
 
-    if (!pData->fClosestream ((mng_handle)pData))
+#ifndef MNG_NO_OPEN_CLOSE_STREAM
+    if (pData->fClosestream && !pData->fClosestream ((mng_handle)pData))
+    {
       MNG_ERROR (pData, MNG_APPIOERROR)
+    }
+#endif
   }
 
   return MNG_NOERROR;
@@ -124,12 +140,145 @@ mng_retcode mng_process_eof (mng_datap pData)
 
 /* ************************************************************************** */
 
+mng_retcode mng_release_pushdata (mng_datap pData)
+{
+  mng_pushdatap pFirst  = pData->pFirstpushdata;
+  mng_pushdatap pNext   = pFirst->pNext;
+
+#ifdef MNG_SUPPORT_TRACE
+  MNG_TRACE (pData, MNG_FN_RELEASE_PUSHDATA, MNG_LC_START)
+#endif
+
+  pData->pFirstpushdata = pNext;       /* next becomes the first */
+
+  if (!pNext)                          /* no next? => no last! */
+    pData->pLastpushdata = MNG_NULL;
+                                       /* buffer owned and release callback defined? */
+  if ((pFirst->bOwned) && (pData->fReleasedata))
+    pData->fReleasedata ((mng_handle)pData, pFirst->pData, pFirst->iLength);
+  else                                 /* otherwise use internal free mechanism */
+    MNG_FREEX (pData, pFirst->pData, pFirst->iLength)
+                                       /* and free it */
+  MNG_FREEX (pData, pFirst, sizeof(mng_pushdata))
+
+#ifdef MNG_SUPPORT_TRACE
+  MNG_TRACE (pData, MNG_FN_RELEASE_PUSHDATA, MNG_LC_END)
+#endif
+
+  return MNG_NOERROR;
+}
+
+/* ************************************************************************** */
+
+mng_retcode mng_release_pushchunk (mng_datap pData)
+{
+  mng_pushdatap pFirst  = pData->pFirstpushchunk;
+  mng_pushdatap pNext   = pFirst->pNext;
+
+#ifdef MNG_SUPPORT_TRACE
+  MNG_TRACE (pData, MNG_FN_RELEASE_PUSHCHUNK, MNG_LC_START)
+#endif
+
+  pData->pFirstpushchunk = pNext;      /* next becomes the first */
+
+  if (!pNext)                          /* no next? => no last! */
+    pData->pLastpushchunk = MNG_NULL;
+                                       /* buffer owned and release callback defined? */
+  if ((pFirst->bOwned) && (pData->fReleasedata))
+    pData->fReleasedata ((mng_handle)pData, pFirst->pData, pFirst->iLength);
+  else                                 /* otherwise use internal free mechanism */
+    MNG_FREEX (pData, pFirst->pData, pFirst->iLength)
+                                       /* and free it */
+  MNG_FREEX (pData, pFirst, sizeof(mng_pushdata))
+
+#ifdef MNG_SUPPORT_TRACE
+  MNG_TRACE (pData, MNG_FN_RELEASE_PUSHCHUNK, MNG_LC_END)
+#endif
+
+  return MNG_NOERROR;
+}
+
+/* ************************************************************************** */
+
+MNG_LOCAL mng_retcode read_data (mng_datap    pData,
+                                 mng_uint8p   pBuf,
+                                 mng_uint32   iSize,
+                                 mng_uint32 * iRead)
+{
+  mng_retcode   iRetcode;
+  mng_uint32    iTempsize = iSize;
+  mng_uint8p    pTempbuf  = pBuf;
+  mng_pushdatap pPush     = pData->pFirstpushdata;
+  mng_uint32    iPushsize = 0;
+  *iRead                  = 0;         /* nothing yet */
+
+#ifdef MNG_SUPPORT_TRACE
+  MNG_TRACE (pData, MNG_FN_READ_DATA, MNG_LC_START)
+#endif
+
+  while (pPush)                        /* calculate size of pushed data */
+  {
+    iPushsize += pPush->iRemaining;
+    pPush      = pPush->pNext;
+  }
+
+  if (iTempsize <= iPushsize)          /* got enough push data? */
+  {
+    while (iTempsize)
+    {
+      pPush = pData->pFirstpushdata;
+                                       /* enough data remaining in this buffer? */
+      if (pPush->iRemaining <= iTempsize)
+      {                                /* no: then copy what we've got */
+        MNG_COPY (pTempbuf, pPush->pDatanext, pPush->iRemaining)
+                                       /* move pointers & lengths */
+        pTempbuf  += pPush->iRemaining;
+        *iRead    += pPush->iRemaining;
+        iTempsize -= pPush->iRemaining;
+                                       /* release the depleted buffer */
+        iRetcode = mng_release_pushdata (pData);
+        if (iRetcode)
+          return iRetcode;
+      }
+      else
+      {                                /* copy the needed bytes */
+        MNG_COPY (pTempbuf, pPush->pDatanext, iTempsize)
+                                       /* move pointers & lengths */
+        pPush->iRemaining -= iTempsize;
+        pPush->pDatanext  += iTempsize;
+        pTempbuf          += iTempsize;
+        *iRead            += iTempsize;
+        iTempsize         = 0;         /* all done!!! */
+      }
+    }
+  }
+  else
+  {
+    mng_uint32 iTempread = 0;
+                                       /* get it from the app then */
+    if (!pData->fReaddata (((mng_handle)pData), pTempbuf, iTempsize, &iTempread))
+      MNG_ERROR (pData, MNG_APPIOERROR);
+
+    *iRead += iTempread;
+  }
+
+#ifdef MNG_SUPPORT_TRACE
+  MNG_TRACE (pData, MNG_FN_READ_DATA, MNG_LC_END)
+#endif
+
+  return MNG_NOERROR;
+}
+
+/* ************************************************************************** */
+
 MNG_LOCAL mng_retcode read_databuffer (mng_datap    pData,
                                        mng_uint8p   pBuf,
                                        mng_uint8p * pBufnext,
                                        mng_uint32   iSize,
                                        mng_uint32 * iRead)
 {
+  mng_retcode iRetcode;
+  
 #ifdef MNG_SUPPORT_TRACE
   MNG_TRACE (pData, MNG_FN_READ_DATABUFFER, MNG_LC_START)
 #endif
@@ -173,13 +322,13 @@ MNG_LOCAL mng_retcode read_databuffer (mng_datap    pData,
                                        /* calculate how much to get */
       iRemain = iSize - (mng_uint32)(*pBufnext - pBuf);
                                        /* let's go get it */
-      if (!pData->fReaddata (((mng_handle)pData), *pBufnext, iRemain, &iTemp))
-        MNG_ERROR (pData, MNG_APPIOERROR);
+      iRetcode = read_data (pData, *pBufnext, iRemain, &iTemp);
+      if (iRetcode)
+        return iRetcode;
                                        /* first read after suspension return 0 means EOF */
       if ((pData->iSuspendpoint) && (iTemp == 0))
       {                                /* that makes it final */
         mng_retcode iRetcode = mng_process_eof (pData);
-
         if (iRetcode)                  /* on error bail out */
           return iRetcode;
                                        /* indicate the source is depleted */
@@ -217,15 +366,15 @@ MNG_LOCAL mng_retcode read_databuffer (mng_datap    pData,
                                        /* now read some more data */
         pTemp = pData->pSuspendbufnext + pData->iSuspendbufleft;
 
-        if (!pData->fReaddata (((mng_handle)pData), pTemp, MNG_SUSPENDREQUESTSIZE, &iTemp))
-          MNG_ERROR (pData, MNG_APPIOERROR);
+        iRetcode = read_data (pData, pTemp, MNG_SUSPENDREQUESTSIZE, &iTemp);
+        if (iRetcode)
+          return iRetcode;
                                        /* adjust fill-counter */
         pData->iSuspendbufleft += iTemp;
                                        /* first read after suspension returning 0 means EOF */
         if ((pData->iSuspendpoint) && (iTemp == 0))
         {                              /* that makes it final */
           mng_retcode iRetcode = mng_process_eof (pData);
-
           if (iRetcode)                /* on error bail out */
             return iRetcode;
 
@@ -243,7 +392,7 @@ MNG_LOCAL mng_retcode read_databuffer (mng_datap    pData,
 
         }
 
-        pData->iSuspendpoint = 0;      /* reset it here in case we loop back */ 
+        pData->iSuspendpoint = 0;      /* reset it here in case we loop back */
       }
 
       if ((!pData->bSuspended) && (!pData->bEOF))
@@ -259,13 +408,11 @@ MNG_LOCAL mng_retcode read_databuffer (mng_datap    pData,
   }
   else
   {
-    if (!pData->fReaddata (((mng_handle)pData), (mng_ptr)pBuf, iSize, iRead))
-    {
-      if (*iRead == 0)                 /* suspension required ? */
-        pData->bSuspended = MNG_TRUE;
-      else
-        MNG_ERROR (pData, MNG_APPIOERROR);
-    }
+    iRetcode = read_data (pData, (mng_ptr)pBuf, iSize, iRead);
+    if (iRetcode)
+      return iRetcode;
+    if (*iRead == 0)                   /* suspension required ? */
+      pData->bSuspended = MNG_TRUE;
   }
 
   pData->iSuspendpoint = 0;            /* safely reset it here ! */
@@ -323,8 +470,10 @@ MNG_LOCAL mng_retcode process_raw_chunk (mng_datap  pData,
     {MNG_UINT_DISC, mng_init_disc, mng_free_disc, mng_read_disc, mng_write_disc, mng_assign_disc, 0, 0},
 #endif
 #ifndef MNG_NO_DELTA_PNG
+#ifndef MNG_SKIPCHUNK_DROP
     {MNG_UINT_DROP, mng_init_drop, mng_free_drop, mng_read_drop, mng_write_drop, mng_assign_drop, 0, 0},
 #endif
+#endif
 #ifndef MNG_SKIPCHUNK_LOOP
     {MNG_UINT_ENDL, mng_init_endl, mng_free_endl, mng_read_endl, mng_write_endl, mng_assign_endl, 0, 0},
 #endif
@@ -529,13 +678,81 @@ MNG_LOCAL mng_retcode process_raw_chunk (mng_datap  pData,
 
 /* ************************************************************************** */
 
+MNG_LOCAL mng_retcode check_chunk_crc (mng_datap  pData,
+                                       mng_uint8p pBuf,
+                                       mng_uint32 iBuflen)
+{
+  mng_uint32  iCrc;                    /* calculated CRC */
+  mng_bool    bDiscard = MNG_FALSE;
+  mng_retcode iRetcode = MNG_NOERROR;
+
+#ifdef MNG_SUPPORT_TRACE
+  MNG_TRACE (pData, MNG_FN_READ_CHUNK_CRC, MNG_LC_START)
+#endif
+
+  if (pData->iCrcmode & MNG_CRC_INPUT) /* crc included ? */
+  {
+    mng_bool bCritical = (mng_bool)((*pBuf & 0x20) == 0);
+    mng_uint32 iL = iBuflen - (mng_uint32)(sizeof (iCrc));
+
+    if (((bCritical ) && (pData->iCrcmode & MNG_CRC_CRITICAL )) ||
+        ((!bCritical) && (pData->iCrcmode & MNG_CRC_ANCILLARY)))
+    {                                  /* calculate the crc */
+      iCrc = mng_crc (pData, pBuf, iL);
+                                       /* and check it */
+      if (!(iCrc == mng_get_uint32 (pBuf + iL)))
+      {
+        mng_bool bWarning = MNG_FALSE;
+        mng_bool bError   = MNG_FALSE;
+
+        if (bCritical)
+        {
+          switch (pData->iCrcmode & MNG_CRC_CRITICAL)
+          {
+            case MNG_CRC_CRITICAL_WARNING  : { bWarning = MNG_TRUE; break; }
+            case MNG_CRC_CRITICAL_ERROR    : { bError   = MNG_TRUE; break; }
+          }
+        }
+        else
+        {
+          switch (pData->iCrcmode & MNG_CRC_ANCILLARY)
+          {
+            case MNG_CRC_ANCILLARY_DISCARD : { bDiscard = MNG_TRUE; break; }
+            case MNG_CRC_ANCILLARY_WARNING : { bWarning = MNG_TRUE; break; }
+            case MNG_CRC_ANCILLARY_ERROR   : { bError   = MNG_TRUE; break; }
+          }
+        }
+
+        if (bWarning)
+          MNG_WARNING (pData, MNG_INVALIDCRC)
+        if (bError)
+          MNG_ERROR (pData, MNG_INVALIDCRC)
+      }
+    }
+
+    if (!bDiscard)                     /* still processing ? */
+      iRetcode = process_raw_chunk (pData, pBuf, iL);
+  }
+  else
+  {                                    /* no crc => straight onto processing */
+    iRetcode = process_raw_chunk (pData, pBuf, iBuflen);
+  }
+
+#ifdef MNG_SUPPORT_TRACE
+  MNG_TRACE (pData, MNG_FN_READ_CHUNK_CRC, MNG_LC_END)
+#endif
+
+  return iRetcode;
+}
+
+/* ************************************************************************** */
+
 MNG_LOCAL mng_retcode read_chunk (mng_datap  pData)
 {
   mng_uint32  iBufmax   = pData->iReadbufsize;
   mng_uint8p  pBuf      = pData->pReadbuf;
   mng_uint32  iBuflen   = 0;           /* number of bytes requested */
   mng_uint32  iRead     = 0;           /* number of bytes read */
-  mng_uint32  iCrc;                    /* calculated CRC */
   mng_retcode iRetcode  = MNG_NOERROR;
 
 #ifdef MNG_SUPPORT_TRACE
@@ -629,81 +846,87 @@ MNG_LOCAL mng_retcode read_chunk (mng_datap  pData)
       if (pData->bSuspended)           /* suspended ? */
         pData->iSuspendpoint = 2;
       else                             /* save the length */
+      {
         pData->iChunklen = mng_get_uint32 (pBuf);
+        if (pData->iChunklen > 0x7ffffff)
+           return MNG_INVALIDLENGTH;
+      }
 
     }
 
     if (!pData->bSuspended)            /* still going ? */
     {                                  /* previously suspended or not eof ? */
       if ((pData->iSuspendpoint > 2) || (iRead == iBuflen))
-      {                                /* determine length chunkname + data + crc */
-        iBuflen = pData->iChunklen + (mng_uint32)(sizeof (mng_chunkid) + sizeof (iCrc));
-
-        if (iBuflen < iBufmax)         /* does it fit in default buffer ? */
-        {                              /* note that we don't use the full size
-                                          so there's always a zero-byte at the
-                                          very end !!! */
-          iRetcode = read_databuffer (pData, pBuf, &pData->pReadbufnext, iBuflen, &iRead);
-
-          if (iRetcode)                /* bail on errors */
-            return iRetcode;
-
-          if (pData->bSuspended)       /* suspended ? */
-            pData->iSuspendpoint = 3;
-          else
-          {
-            if (iRead != iBuflen)      /* did we get all the data ? */
-              iRetcode = MNG_UNEXPECTEDEOF;
-            else
-            {
-              mng_uint32 iL = iBuflen - (mng_uint32)(sizeof (iCrc));
-                                       /* calculate the crc */
-              iCrc = mng_crc (pData, pBuf, iL);
-                                       /* and check it */
-              if (!(iCrc == mng_get_uint32 (pBuf + iL)))
-                iRetcode = MNG_INVALIDCRC;
-              else
-                iRetcode = process_raw_chunk (pData, pBuf, iL);
-            }
-          }
-        }
+      {                                /* determine length chunkname + data (+ crc) */
+        if (pData->iCrcmode & MNG_CRC_INPUT)
+          iBuflen = pData->iChunklen + (mng_uint32)(sizeof (mng_chunkid) + sizeof (mng_uint32));
         else
+          iBuflen = pData->iChunklen + (mng_uint32)(sizeof (mng_chunkid));
+
+                                       /* do we have enough data in the current push buffer ? */
+        if ((pData->pFirstpushdata) && (iBuflen <= pData->pFirstpushdata->iRemaining))
         {
-          if (iBuflen > 16777216)      /* is the length incredible? */
-            MNG_ERROR (pData, MNG_IMPROBABLELENGTH);
-
-          if (!pData->iSuspendpoint)   /* create additional large buffer ? */
-          {                            /* again reserve space for the last zero-byte */
-            pData->iLargebufsize = iBuflen + 1;
-            pData->pLargebufnext = MNG_NULL;
-            MNG_ALLOC (pData, pData->pLargebuf, pData->iLargebufsize)
-          }
-
-          iRetcode = read_databuffer (pData, pData->pLargebuf, &pData->pLargebufnext, iBuflen, &iRead);
+          mng_pushdatap pPush  = pData->pFirstpushdata;
+          pBuf                 = pPush->pDatanext;
+          pPush->pDatanext    += iBuflen;
+          pPush->iRemaining   -= iBuflen;
+          pData->iSuspendpoint = 0;    /* safely reset this here ! */
 
+          iRetcode = check_chunk_crc (pData, pBuf, iBuflen);
           if (iRetcode)
             return iRetcode;
 
-          if (pData->bSuspended)       /* suspended ? */
-            pData->iSuspendpoint = 4;
-          else
-          {
-            if (iRead != iBuflen)      /* did we get all the data ? */
-              iRetcode = MNG_UNEXPECTEDEOF;
+          if (!pPush->iRemaining)      /* buffer depleted? then release it */
+            iRetcode = mng_release_pushdata (pData);
+        }
+        else
+        {
+          if (iBuflen < iBufmax)       /* does it fit in default buffer ? */
+          {                            /* note that we don't use the full size
+                                          so there's always a zero-byte at the
+                                          very end !!! */
+            iRetcode = read_databuffer (pData, pBuf, &pData->pReadbufnext, iBuflen, &iRead);
+            if (iRetcode)              /* bail on errors */
+              return iRetcode;
+
+            if (pData->bSuspended)     /* suspended ? */
+              pData->iSuspendpoint = 3;
             else
             {
-              mng_uint32 iL = iBuflen - (mng_uint32)(sizeof (iCrc));
-                                       /* calculate the crc */
-              iCrc = mng_crc (pData, pData->pLargebuf, iL);
-                                       /* and check it */
-              if (!(iCrc == mng_get_uint32 (pData->pLargebuf + iL)))
-                iRetcode = MNG_INVALIDCRC;
+              if (iRead != iBuflen)    /* did we get all the data ? */
+                MNG_ERROR (pData, MNG_UNEXPECTEDEOF)
               else
-                iRetcode = process_raw_chunk (pData, pData->pLargebuf, iL);
+                iRetcode = check_chunk_crc (pData, pBuf, iBuflen);
             }
+          }
+          else
+          {
+            if (iBuflen > 16777216)    /* is the length incredible? */
+              MNG_ERROR (pData, MNG_IMPROBABLELENGTH)
+
+            if (!pData->iSuspendpoint) /* create additional large buffer ? */
+            {                          /* again reserve space for the last zero-byte */
+              pData->iLargebufsize = iBuflen + 1;
+              pData->pLargebufnext = MNG_NULL;
+              MNG_ALLOC (pData, pData->pLargebuf, pData->iLargebufsize)
+            }
+
+            iRetcode = read_databuffer (pData, pData->pLargebuf, &pData->pLargebufnext, iBuflen, &iRead);
+            if (iRetcode)
+              return iRetcode;
+
+            if (pData->bSuspended)     /* suspended ? */
+              pData->iSuspendpoint = 4;
+            else
+            {
+              if (iRead != iBuflen)    /* did we get all the data ? */
+                MNG_ERROR (pData, MNG_UNEXPECTEDEOF)
+              else
+                iRetcode = check_chunk_crc (pData, pData->pLargebuf, iBuflen);
                                        /* cleanup additional large buffer */
-            MNG_FREE (pData, pData->pLargebuf, pData->iLargebufsize)
-          }  
+              MNG_FREE (pData, pData->pLargebuf, pData->iLargebufsize)
+            }
+          }
         }
 
         if (iRetcode)                  /* on error bail out */
@@ -723,7 +946,7 @@ MNG_LOCAL mng_retcode read_chunk (mng_datap  pData)
 #else
             (pData->bHasIHDR || pData->bHasMHDR))
 #endif
-          MNG_ERROR (pData, MNG_UNEXPECTEDEOF);
+          MNG_ERROR (pData, MNG_UNEXPECTEDEOF)
       } 
     }
   }
@@ -747,6 +970,20 @@ MNG_LOCAL mng_retcode read_chunk (mng_datap  pData)
 
 /* ************************************************************************** */
 
+MNG_LOCAL mng_retcode process_pushedchunk (mng_datap pData)
+{
+  mng_retcode   iRetcode;
+  mng_pushdatap pPush = pData->pFirstpushchunk;
+
+  iRetcode = process_raw_chunk (pData, pPush->pData, pPush->iLength);
+  if (iRetcode)
+    return iRetcode;
+
+  return mng_release_pushchunk (pData);
+}
+
+/* ************************************************************************** */
+
 mng_retcode mng_read_graphic (mng_datap pData)
 {
   mng_uint32  iBuflen;                 /* number of bytes requested */
@@ -806,7 +1043,10 @@ mng_retcode mng_read_graphic (mng_datap pData)
       if ((pData->bReading) && (!pData->bDisplaying))
         pData->bTimerset = MNG_FALSE;
 
-      iRetcode = read_chunk (pData);   /* process a chunk */
+      if (pData->pFirstpushchunk)      /* chunks pushed ? */
+        iRetcode = process_pushedchunk (pData); /* process the pushed chunk */
+      else
+        iRetcode = read_chunk (pData); /* read & process a chunk */
 
       if (iRetcode)                    /* on error bail out */
         return iRetcode;
